// 2019-08-02 ffgl@subpixels.com
// - Make inline shims for Compile(), FindUniform(), Set() and BindTexture() with const string& arguments
// - Add variants of Set() with uniformLocation arguments
// - Replace CompileVertexShader(), CompileGeometryShader(), CompileFragmentShader() by generic CompileShader()
// 2019-11-20 ffgl@subpixels.com
// - Add class Uniform
// - Add Locate()

#pragma once

#include <ffgl/FFGL.h>// For OpenGL
#include <vector>
#include <string>
#include <map>
#include <string>

namespace ffglex
{
/**
 * The Shader is a helper class you can use to compile glsl shader strings
 * and link them together into a shader program. You can then get the program id
 * and use it to bind the resulting shader to use it for the next rendering operation.
 */
class Shader final
{
public:
	typedef struct Uniform
	{
	public:
		std::string name;
		GLint location;

		Uniform( const std::string& name ) :
			name( name ),
			location( -1 ){};
		~Uniform(){};

		//TODO: ffgl@subpixels.com - Refactor Uniform.Set(float f) as Shader.Set(Uniform u, float f) { if(u.location != -1) glUniform1f( location, f ); }

		// Set float
		inline void Set( float f )
		{
			if( location != -1 )
				glUniform1f( location, f );
		}

		// Set vec2
		inline void Set( float f1, float f2 )
		{
			if( location != -1 )
				glUniform2f( location, f1, f2 );
		}

		// Set vec3
		inline void Set( float f1, float v2, float f3 )
		{
			if( location != -1 )
				glUniform3f( location, f1, v2, f3 );
		}

		// Set vec4
		inline void Set( float f1, float f2, float f3, float f4 )
		{
			if( location != -1 )
				glUniform4f( location, f1, f2, f3, f4 );
		}

		// Set boolean
		inline void Setb( bool b )
		{
			// Note: can't pass a bool directly; need to pass as an int
			if( location != -1 )
				glUniform1i( location, static_cast< int >( b ) );
		}

		// Set bvec2
		inline void Setb( bool b1, bool b2 )
		{
			// Note: can't pass bools directly; need to pass as ints
			if( location != -1 )
				glUniform2i( location, static_cast< int >( b1 ), static_cast< int >( b2 ) );
		}

		// Set bvec3
		inline void Setb( bool b1, bool b2, bool b3 )
		{
			// Note: can't pass bools directly; need to pass as ints
			if( location != -1 )
				glUniform3i( location, static_cast< int >( b1 ), static_cast< int >( b2 ), static_cast< int >( b3 ) );
		}

		// Set bvec4
		inline void Setb( bool b1, bool b2, bool b3, bool b4 )
		{
			// Note: can't pass bools directly; need to pass as ints
			if( location != -1 )
				glUniform4i( location, static_cast< int >( b1 ), static_cast< int >( b2 ), static_cast< int >( b3 ), static_cast< int >( b4 ) );
		}

		// Set integer
		inline void Seti( int i )
		{
			if( location != -1 )
				glUniform1i( location, i );
		}

		// Set ivec2
		inline void Seti( int i1, int i2 )
		{
			if( location != -1 )
				glUniform2i( location, i1, i2 );
		}

		// Set ivec3
		inline void Seti( int i1, int i2, int i3 )
		{
			if( location != -1 )
				glUniform3i( location, i1, i2, i3 );
		}

		// Set ivec4
		inline void Seti( GLint location, int i1, int i2, int i3, int i4 )
		{
			if( location != -1 )
				glUniform4i( location, i1, i2, i3, i4 );
		}
	} Uniform;

	Shader(); //The default constructor just initializes this object to represent no shader.
	~Shader();//Destructing this object just checks and lets you know if you've properly released the GL resources.

	void AddTransformFeedbackVarying( const std::string& varyingName );//Add a varying that will be captured using a transform feedback.
	bool Compile( const std::string& vertexShader, const std::string& geometryShader, const std::string& fragmentShader );
	inline bool Compile( const std::string& vertexShader, const std::string& fragmentShader )
	{
		return Compile( vertexShader, {}, fragmentShader );
	}

	void FreeGLResources();//Frees any GL resources that this shader is holding.

	bool IsReady() const;                              //Whether or not previous complilation succeeded and this shader is ready to be used for rendering.
	void Use() const;                                  //Call glUseProgram with the current programID
	GLuint GetGLID() const;                            //Gets the OpenGL ID that represents the linked shader program.
	GLint FindUniform( const std::string& name ) const;//Finds the uniform location that needs to be used to addess a uniform with a certain name.

	/// Locate a uniform (sets uniform.location); returns true if located, false otherwise
	inline bool Locate( Uniform& uniform ) const
	{
		uniform.location = FindUniform( uniform.name );
		return ( uniform.location != -1 );
	}

	void Set( const std::string& name, float value );
	void Set( const std::string& name, float v1, float v2 );
	void Set( const std::string& name, float v1, float v2, float v3 );
	void Set( const std::string& name, float v1, float v2, float v3, float v4 );
	void Set( const std::string& name, int value );

	// Set float
	inline void Set(GLint uniformLocation, float f)
	{
		glUniform1f( uniformLocation, f );
	}

	// Set vec2
	inline void Set( GLint uniformLocation, float f1, float f2 )
	{
		glUniform2f( uniformLocation, f1, f2 );
	}

	// Set vec3
	inline void Set( GLint uniformLocation, float f1, float v2, float f3 )
	{
		glUniform3f( uniformLocation, f1, v2, f3 );
	}

	// Set vec4
	inline void Set( GLint uniformLocation, float f1, float f2, float f3, float f4 )
	{
		glUniform4f( uniformLocation, f1, f2, f3, f4 );
	}

	// Set boolean
	inline void Setb( GLint uniformLocation, bool b )
	{
		// Note: can't pass a bool directly; need to pass as an int
		glUniform1i( uniformLocation, static_cast< int >( b ) );
	}

	// Set bvec2
	inline void Setb( GLint uniformLocation, bool b1, bool b2 )
	{
		// Note: can't pass bools directly; need to pass as ints
		glUniform2i( uniformLocation, static_cast< int >( b1 ), static_cast< int >( b2 ) );
	}

	// Set bvec3
	inline void Setb( GLint uniformLocation, bool b1, bool b2, bool b3 )
	{
		// Note: can't pass bools directly; need to pass as ints
		glUniform3i( uniformLocation, static_cast< int >( b1 ), static_cast< int >( b2 ), static_cast< int >( b3 ) );
	}

	// Set bvec4
	inline void Setb( GLint uniformLocation, bool b1, bool b2, bool b3, bool b4 )
	{
		// Note: can't pass bools directly; need to pass as ints
		glUniform4i( uniformLocation, static_cast< int >( b1 ), static_cast< int >( b2 ), static_cast< int >( b3 ), static_cast< int >( b4 ) );
	}

	// Set integer
	inline void Seti( GLint uniformLocation, int i )
	{
		glUniform1i( uniformLocation, i );
	}

	// Set ivec2
	inline void Seti( GLint uniformLocation, int i1, int i2 )
	{
		glUniform2i( uniformLocation, i1, i2 );
	}

	// Set ivec3
	inline void Seti( GLint uniformLocation, int i1, int i2, int i3 )
	{
		glUniform3i( uniformLocation, i1, i2, i3 );
	}

	// Set ivec4
	inline void Seti( GLint uniformLocation, int i1, int i2, int i3, int i4 )
	{
		glUniform4i( uniformLocation, i1, i2, i3, i4 );
	}

	void BindTexture( const std::string& name, int texture, const ffgl::TextureStruct& fbo );

	Shader& operator=( const Shader& ) = delete;
	Shader& operator=( Shader&& ) = delete;

private:
	//Internal utility function that tries to compile a shader. Doesn't clean up on failure.
	bool CompileShader( const GLuint shaderID, const std::string& shaderCode, const std::string& shaderName );

	//Internal utility function that tries to link the two previously compiled vertex and fragment shaders into a program. Doesn't clean up on failure.
	bool LinkProgram();

	GLuint vertexShaderID;  //!< The ID OpenGL gave our vertex shader. 0 for invalid.
	GLuint geometryShaderID;//!< The ID OpenGL gave our geometry shader. 0 if not present or invalid.
	GLuint fragmentShaderID;//!< The ID OpenGL gave our fragment shader. 0 for invalid.
	GLuint programID;       //!< The ID OpenGL gave our shader program. BindTexture this to use this shader. 0 for invalid.

	std::vector< std::string > transformFeedbackVaryings;//!< The varyings that will be captured using a transform feedback. Ordered in the order of capturing.
};

}//End namespace ffglex